
/*
	import.c - Raw and ASCII dtat import
*/


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "tiffstuff.h"
#include "import.h"






void swapme000 (char* p)
{
char b;
	
	b= p[0]; p[0]=p[1]; p[1]=b;
}

unsigned short swapme001 (unsigned short x)
{
	return (((x >> 8) & 0x00ff) | ((x << 8) & 0xff00));
}


#define SWAPME(X)	((((X) >> 8) & 0x00ff) | (((X) << 8) & 0xff00))

/********************************************************

	do_import reads non-ascii raw formats

*********************************************************/


int do_import (image_type *img, import_params params)
{
FILE* FP;
char s[1024];
long bytesread,i,idx; 
int esize, warn;
unsigned short* p; /* should be unsigned to avoid setting of leading bits in negative numbers */


	warn=0;
	FP = fopen (params.filename, "rb");
	if (!FP)
	{
		strcpy (s, "Error: Can't open file\n");
		strcat (s, params.filename);
		ShowMessage (s);
	}
	else
	{
		statusbar_display ("Importing file...");
		if (allocate_image (img, params.imgtype, params.xdim,params.ydim,params.zdim)==-1) /* a 3D image */
		{
			fatal_memerr ();
		}
		
		/* Skip the offset. We can use fseek to do this */
		
		if (params.ofs)
		{
			if (!fseek (FP, params.ofs, SEEK_SET))
				dp (1, "Import: Skipped %li bytes\n", params.ofs);
			else
				dp (1, "ERROR: Could not skip %li bytes offset\n", params.ofs);
		}
		dp (1, "Import: Reading data chunk, dimensions (%li x %li x %li)\n",
				params.xdim,params.ydim,params.zdim);


		/* Handle iB_DOUBLE separately, because we need to truncate precision */
		
		if (params.src_type==iB_DOUBLE)
		{
			float *pf = mainimg.data;
			double buf;
			esize = sizeof(double);
			bytesread = 0;
			for (idx=0; idx < params.xdim*params.ydim*params.zdim; idx++)
			{
				bytesread += fread (&buf, esize, 1, FP);
				pf[idx] = (float)buf;
			}
		
		}
		else		/* Otherwise we can read it in one big chunk */
		{

			/* Actually, "bytesread" is really "elements read" */
			esize = DATALEN(params.imgtype);
			bytesread = fread (img->data, esize, params.xdim*params.ydim*params.zdim, FP);
		}

		fclose (FP);

		if (bytesread != (params.xdim*params.ydim*params.zdim))
		{
			warn=2;
			dp (1, "Warning: %d elements requested but %d elements read\n",params.xdim*params.ydim*params.zdim, bytesread);
		}
		else
			dp (1, "Import: %d elements read -> OK\n",bytesread);

		
		/* If this is a short type, handle little-endian if requested */
		
		if ((params.imgtype==B_SHORT) && (params.imp_swab))
		{
			p = img->data;
			dp (1, "Import: Swapping msb/lsb for little-endian\n");
			for (i=0; i<params.xdim*params.ydim*params.zdim; i++)
			{
				p[i] = SWAPME(p[i]);
			}
		}

		/* No way around a sweep over the data - we need to determine min and max */

		make_minmax (img);
		return warn;
		
	
	}
	return -1;
	
}



/********************************************************

	do_import_ascii reads ascii raw formats

*********************************************************/


int do_import_ascii (image_type *img, import_params params)
{
FILE* FP;
char s[1024];
long bytesread,i; 
short* sp;
float* fp;
char* cp;
int esize;
float f; 
int si;
int warn;




	FP = fopen (params.filename, "rt");
	if (!FP)
	{
		strcpy (s, "Error: Can't open file\n");
		strcat (s, params.filename);
		ShowMessage (s);
	}
	else
	{
		statusbar_display ("Importing file...");
		if (allocate_image (img, params.imgtype, params.xdim,params.ydim,params.zdim)==-1) /* a 3D image */
		{
			fatal_memerr ();
		}
		
		sp = img->data; fp = img->data; cp = img->data;
		
		esize=1; if (params.imgtype==B_RGB) esize=3;
		i=0; bytesread=0; warn=0;
		while ((i<params.xdim*params.ydim*params.zdim*esize) && (!feof(FP)))
		{
			switch (params.imgtype)
			{
				case B_CHAR:
					fscanf (FP, "%d",&si);
					if ((si<0) || (si>255)) warn=1;
					cp[i] = si;
					break;
				case B_SHORT:
					fscanf (FP, "%d",&si);
					if ((si<-32768) || (si>32767)) warn=1;
					sp[i] = si;
					break;
				case B_FLOAT:
					fscanf (FP, "%f",&f);
					fp[i] = f;
					break;
				case B_RGB:
					fscanf (FP, "%d",&si);
					if ((si<0) || (si>255)) warn=1;
					cp[i] = si;
					break;
			}
			bytesread++; i++;
		}

		fclose (FP);
		
		if (bytesread!=params.xdim*params.ydim*params.zdim*esize) warn=2;

		/* No way around a sweep over the data - we need to determine min and max */

		make_minmax (img);
		return warn;
	
				
	
	}
	return -1;
	
}



void import_mainimg (import_params params)
{
int err, esize;
char s[1024];

	if (file_ok) close_file();

	if (params.is_ascii)
		dp (2, "Importing ASCII file ");
	else
		dp (2, "Importing raw (binary) file ");

	dp (2, " with dimensions (%d,%d,%d)\n",params.xdim,params.ydim,params.zdim);

	if (params.is_ascii)
		err=do_import_ascii (&mainimg, params);
	else
		err=do_import(&mainimg, params);

	esize=0;
	if (!err)
	{

		/* Display image */
		prepare_display_window (params.filename);
		/* Finally, create an imlib image for display */
		dp (1, "Import: Done reading. Preparing imlib buffer\n");
		make_imlib_buffer_from_main_image (0);
		file_ok = 1;
		esize = DATALEN (mainimg.imgtype);
	}
	if (err==2)
		sprintf (s, "Import problem: Data values do not match size");
	else if (err==1)
		sprintf (s, "Import problem: Value range overflow");
	else 
		sprintf (s, "File imported OK. Size = %li MB", 
				params.xdim*params.ydim*params.zdim*esize/1024576);

	redraw (0);
	statusbar_display (s);
}



/*===================== Import format dialog =============================*/

TYPE_SELECTOR	imgtype;
SRC_TYPE src_type;
long xdim, ydim, zdim;
long ofs;
char infile[1024];
char imp_swab=0;		/* flag whether to swap msb/lsb in shorts */

int inpf_rtncode;
int ascii_import;
  GtkWidget *spinbutton1;
  GtkWidget *spinbutton2;
  GtkWidget *spinbutton3;
  GtkWidget *spinbutton4;
  GtkWidget *imp_types;
  GtkWidget *imp_types_menu;
  GtkWidget* imp_totalsize; /* Display total file size */
  GtkWidget *check_little_endian;


char size_range (long size, long guess)
{
long n, l1, l2;

	n = size/guess;		/* number of slices */
	l1 = n*guess;		/* truncated */
	l2 = l1 + l1/10;	/* upper tolerance limit */
	if ((size > l1) && (size <= l2)) return 1;
	return 0;

}


void guess_size (char* fn, int* xs, int* ys, int* zs, int* ofs, TYPE_SELECTOR* t)
{
struct stat statresult;		/* to get file size */
long size,n;

	if (stat (fn, &statresult))	/* error if nonzero */
	{
		dp (1, "Import: STAT failed\n");
		return;	/* Don't modify the defaults */
	}
	
	size = statresult.st_size;
	
	/* test 1: DCV-1300 file */
	
	n = size / 2686240;
	if (n*2686240 == size)
	{
		*xs = 1304; *ys = 1030; *zs=n; *ofs=0; *t=B_SHORT;
		return;
	}
	n = size / 1343120;
	if (n*1343120 == size)
	{
		*xs = 1304; *ys = 1030; *zs=n; *ofs=0; *t=B_CHAR;
		return;
	}
	
	/* test 2: 256x256 and 512x512 no-offset raw files */
	
	n = size/524288; /* MUST start with 16-bit 512x512 */
	if (n*524288 == size)
	{
		*xs=*ys=512; *zs=n; *ofs=0; *t=B_SHORT;
		return;
	}
	n = size/262144; 
	if (n*262144 == size)
	{
		*xs=*ys=512; *zs=n; *ofs=0; *t=B_CHAR;
		return;
	}
	n = size/131072; 
	if (n*131072 == size)
	{
		*xs=*ys=256; *zs=n; *ofs=0; *t=B_SHORT;
		return;
	}
	n = size/65536; 
	if (n*65536 == size)
	{
		*xs=*ys=256; *zs=n; *ofs=0; *t=B_CHAR;
		return;
	}
	
	/* Test 3: Check if a "good" number in X and any number in Y yields zero offset */

	n = size/2048;	/* x=512, 4-byte FLOAT */
	if (n*2048==size)
	{
		*xs=512; *ys=n; *t=B_FLOAT;
		return;
	}
	n = size/1536;	/* x=512, RGB */
	if (n*1536==size)
	{
		*xs=512; *ys=n; *t=B_RGB;
		return;
	}
	n = size/1024;	/* x=256, 4-byte FLOAT _or_ x=512, 2-byte INT */
	if (n*2048==size)
	{
		*xs=512; *ys=n; *t=B_SHORT;
		return;
	}
	n = size/768;	/* x=256, RGB */
	if (n*1536==size)
	{
		*xs=256; *ys=n; *t=B_RGB;
		return;
	}

	/* Test 4: Check if we are within 10% of one of the above */
	
	if (size_range(size,524288)) /* MUST start with 16-bit 512x512 */
	{
		*xs=*ys=512; *zs=size/524288; 
		*ofs=size - *zs *524288 ; *t=B_SHORT;
		return;
	}
	if (size_range(size,262144))
	{
		*xs=*ys=512; *zs=size/262144; 
		*ofs=size - *zs * 262144; *t=B_CHAR;
		return;
	}
	if (size_range (size,131072))
	{
		*xs=*ys=256; *zs=size/131072; 
		*ofs=size - *zs * 131072; *t=B_SHORT;
		return;
	}
	if (size_range(size,65536))
	{
		*xs=*ys=256; *zs=size/65536;
		*ofs=size - *zs*65536; *t=B_CHAR;
		return;
	}
	

}


void on_impdlg_ok_clicked (GtkButton *button, gpointer user_data)
{
long totalsize;
char s[20];
import_params params;

	strncpy (params.filename, infile, 1023);
	params.xdim = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton1);
	params.ydim = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton2);
	params.zdim = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton3);
	params.ofs  = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton4);

	params.imp_swab = check_little_endian->state;
	
	totalsize = params.xdim*params.ydim*params.zdim;
	if (src_type==iB_SHORT) totalsize *= sizeof (short);
	else if (src_type==iB_RGB) totalsize *= sizeof (rgbtriplet);
	else if (src_type==iB_FLOAT) totalsize *= sizeof (float);
	else if (src_type==iB_DOUBLE) totalsize *= sizeof (double);

	totalsize += ofs;
	sprintf (s, "%li",totalsize);
	gtk_label_set_text (GTK_LABEL(imp_totalsize), s);

	params.src_type = src_type;
	params.imgtype = imgtype;
	params.is_ascii = ascii_import;

	import_mainimg (params);

   	gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}


void on_impdlg_cancel_clicked (GtkButton *button, gpointer user_data)
{
   gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}

void imp_display_new_size (GtkWidget* widget, gpointer user_data)
{
long totalsize;
long xdim0,ydim0,zdim0,ofs0;
char s[200];

	xdim0 = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton1);
	ydim0 = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton2);
	zdim0 = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton3);
	ofs0  = gtk_spin_button_get_value_as_int ((GtkSpinButton*)spinbutton4);
	
	totalsize = xdim0*ydim0*zdim0;

	if (ascii_import)
	{
		sprintf (s, "%li ASCII values", totalsize);
	}
	else
	{
		if (src_type==iB_SHORT) totalsize *= sizeof (short);
		else if (src_type==iB_RGB) totalsize *= sizeof (rgbtriplet);
		else if (src_type==iB_FLOAT) totalsize *= sizeof (float);
		else if (src_type==iB_DOUBLE) totalsize *= sizeof (double);
		totalsize += ofs0;
		sprintf (s, "%li bytes",totalsize);
	}
	gtk_label_set_text (GTK_LABEL(imp_totalsize), s);
}


void imp_optionmenu_handler (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	
	if (i==0) imgtype=B_CHAR;
	else if (i==1) imgtype=B_SHORT;
	else if (i==2) imgtype=B_RGB;
	else imgtype=B_FLOAT;
	
	if (i==0) src_type=iB_CHAR;
	else if (i==1) src_type=iB_SHORT;
	else if (i==2) src_type=iB_RGB;
	else if (i==3) src_type=iB_FLOAT;
	else if (i==4) src_type=iB_DOUBLE;
	
	imp_display_new_size (NULL,NULL);
	
}

void on_import_ascii_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	ascii_import = !ascii_import;
	imp_display_new_size (NULL, NULL);
}


GtkWidget* create_importdlg (int xs, int ys, int zs, int ofs, TYPE_SELECTOR t)
{
  GtkWidget *importdlg;
  GtkWidget *dialog_vbox2;
  GtkWidget *table1;
  GtkWidget *label1;
  GtkWidget *label2;
  GtkWidget *label3;
  GtkWidget *label4;
  GtkWidget *label5;
  GtkWidget *label6;
  GtkWidget *label8;
  GtkWidget *label9;
  GtkWidget *checkbutton1;
  GtkObject *spinbutton1_adj;
  GtkObject *spinbutton2_adj;
  GtkObject *spinbutton3_adj;
  GtkObject *spinbutton4_adj;
  GtkWidget *imp_types;
  GtkWidget *imp_types_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *dialog_action_area2;
  GtkWidget *hbox2;
  GtkWidget *impdlg_ok;
  GtkWidget *impdlg_cancel;

  importdlg = gtk_dialog_new ();
  gtk_object_set_data (GTK_OBJECT (importdlg), "importdlg", importdlg);
  gtk_window_set_title (GTK_WINDOW (importdlg), "Import Parameters");
  gtk_window_set_policy (GTK_WINDOW (importdlg), TRUE, TRUE, FALSE);

  dialog_vbox2 = GTK_DIALOG (importdlg)->vbox;
  gtk_object_set_data (GTK_OBJECT (importdlg), "dialog_vbox2", dialog_vbox2);
  gtk_widget_show (dialog_vbox2);

  table1 = gtk_table_new (8, 2, FALSE);
  gtk_widget_ref (table1);
  gtk_object_set_data_full (GTK_OBJECT (importdlg), "table1", table1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (table1);
  gtk_box_pack_start (GTK_BOX (dialog_vbox2), table1, TRUE, TRUE, 0);

  label1 = MakeLabel (importdlg,table1,"File type","label1",0,0);
  label2 = MakeLabel (importdlg,table1,"X size","label2",0,1);
  label3 = MakeLabel (importdlg,table1,"Y size","label3",0,2);
  label4 = MakeLabel (importdlg,table1,"Z size","label4",0,3);
  label5 = MakeLabel (importdlg,table1,"Offset","label5",0,4);
  label8 = MakeLabel (importdlg,table1,"Little-endian","label8",0,5);
  label9 = MakeLabel (importdlg,table1,"Import ASCII","label9",0,6);
  label6 = MakeLabel (importdlg,table1,"Total size","label6",0,7);
  /* label7 = MakeLabel (importdlg,table1,"0","label7",1,6); */
  imp_totalsize = MakeLabel (importdlg,table1, "", "imp_totalsize",1,7);

  check_little_endian = gtk_check_button_new_with_label ("(short only)");
  gtk_widget_ref (check_little_endian);
  gtk_object_set_data_full (GTK_OBJECT (window1), "check_little_endian", check_little_endian,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (check_little_endian);
  gtk_table_attach (GTK_TABLE (table1), check_little_endian, 1, 2, 5, 6,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

	/* Parameters for gtk_adjustment_new: Value - lower - upper - stepinc - pageinc - pagesize */

  spinbutton1_adj = gtk_adjustment_new (xs, 1, 65535, 1, 10, 10);
  spinbutton1 = MakeSpinbutton (GTK_ADJUSTMENT(spinbutton1_adj), importdlg, table1, "spinbutton1", 1,1);

  spinbutton2_adj = gtk_adjustment_new (ys, 1, 65535, 1, 10, 10);
  spinbutton2 = MakeSpinbutton (GTK_ADJUSTMENT(spinbutton2_adj), importdlg, table1, "spinbutton2", 1,2);

  spinbutton3_adj = gtk_adjustment_new (zs, 1, 1024, 1, 10, 10);
  spinbutton3 = MakeSpinbutton (GTK_ADJUSTMENT(spinbutton3_adj), importdlg, table1, "spinbutton3", 1,3);

  spinbutton4_adj = gtk_adjustment_new (ofs, 0, 209715200, 1, 10, 10);
  spinbutton4 = MakeSpinbutton (GTK_ADJUSTMENT(spinbutton4_adj), importdlg, table1, "spinbutton4", 1,4);

  checkbutton1 = Checkbutton_in_table (importdlg,table1," ","ascii_chkbtn", 1,6,on_import_ascii_toggled);
  ascii_import = 0;

  imp_types = gtk_option_menu_new ();
  gtk_widget_ref (imp_types);
  gtk_object_set_data_full (GTK_OBJECT (importdlg), "imp_types", imp_types,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (imp_types);
  gtk_table_attach (GTK_TABLE (table1), imp_types, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_container_set_border_width (GTK_CONTAINER (imp_types), 2);
  imp_types_menu = gtk_menu_new ();
  glade_menuitem = Option_menuitem (imp_types_menu,GTK_SIGNAL_FUNC(imp_optionmenu_handler), "8-Bit Grayscale",(gpointer) &menuopt0);
  glade_menuitem = Option_menuitem (imp_types_menu,GTK_SIGNAL_FUNC(imp_optionmenu_handler), "16-Bit Grayscale",(gpointer) &menuopt1);
  glade_menuitem = Option_menuitem (imp_types_menu,GTK_SIGNAL_FUNC(imp_optionmenu_handler), "24-Bit RGB",(gpointer) &menuopt2);
  glade_menuitem = Option_menuitem (imp_types_menu,GTK_SIGNAL_FUNC(imp_optionmenu_handler), "IEEE Float Grayscale",(gpointer) &menuopt3);
  glade_menuitem = Option_menuitem (imp_types_menu,GTK_SIGNAL_FUNC(imp_optionmenu_handler), "IEEE Double Grayscale",(gpointer) &menuopt4);

  gtk_option_menu_set_menu (GTK_OPTION_MENU (imp_types), imp_types_menu);
  gtk_option_menu_set_history (GTK_OPTION_MENU(imp_types), (int)t); /* This typecast just doesn't make any sense, but it works w/o warning */
  imgtype = t;		/* Must simulate imp_optionmenu_handler call */

  dialog_action_area2 = GTK_DIALOG (importdlg)->action_area;
  gtk_object_set_data (GTK_OBJECT (importdlg), "dialog_action_area2", dialog_action_area2);
  gtk_widget_show (dialog_action_area2);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area2), 10);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (importdlg), "hbox2", hbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (dialog_action_area2), hbox2, TRUE, TRUE, 0);

  impdlg_ok = gtk_button_new_with_label ("OK");
  gtk_widget_ref (impdlg_ok);
  gtk_object_set_data_full (GTK_OBJECT (importdlg), "impdlg_ok", impdlg_ok,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (impdlg_ok);
  gtk_box_pack_start (GTK_BOX (hbox2), impdlg_ok, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (impdlg_ok, GTK_CAN_DEFAULT);

  impdlg_cancel = gtk_button_new_with_label ("Cancel");
  gtk_widget_ref (impdlg_cancel);
  gtk_object_set_data_full (GTK_OBJECT (importdlg), "impdlg_cancel", impdlg_cancel,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (impdlg_cancel);
  gtk_box_pack_start (GTK_BOX (hbox2), impdlg_cancel, TRUE, TRUE, 0);

  gtk_signal_connect (GTK_OBJECT (impdlg_ok), "clicked",
                      GTK_SIGNAL_FUNC (on_impdlg_ok_clicked),
                      importdlg);
  gtk_signal_connect (GTK_OBJECT (impdlg_cancel), "clicked",
                      GTK_SIGNAL_FUNC (on_impdlg_cancel_clicked),
                      importdlg);
  gtk_signal_connect (GTK_OBJECT (importdlg), "destroy",
                      GTK_SIGNAL_FUNC (on_impdlg_cancel_clicked),
                      importdlg);
  gtk_signal_connect (GTK_OBJECT (spinbutton1_adj), "value_changed",
                      GTK_SIGNAL_FUNC (imp_display_new_size),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (spinbutton2_adj), "value_changed",
                      GTK_SIGNAL_FUNC (imp_display_new_size),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (spinbutton3_adj), "value_changed",
                      GTK_SIGNAL_FUNC (imp_display_new_size),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (spinbutton4_adj), "value_changed",
                      GTK_SIGNAL_FUNC (imp_display_new_size),
                      NULL);

  imp_display_new_size (NULL, NULL);
  return importdlg;
}



int import_dialog (char* infilename)
{
GtkWidget *importdlg;
int xdef, ydef, zdef, ofsdef;	/* Default settings */
TYPE_SELECTOR tdef;

	strcpy (infile, infilename);
	inpf_rtncode=0;
 
 	xdef=256; ydef=256; zdef=1; ofsdef=0; tdef=0;
	guess_size (infilename, &xdef, &ydef, &zdef, &ofsdef, &tdef);
 
  	importdlg = create_importdlg (xdef,ydef,zdef,ofsdef, tdef);
  	gtk_widget_show (importdlg);

  	return inpf_rtncode;
  
  
}



/*============================================================*/


/*

	Import as RGB: We need up to 3 images here which are 
	put into the three RGB channels.
	Whatever the type, it will be compressed to 8-bit unsigned
*/


GtkWidget *rgbload_redentry;
GtkWidget *rgbload_greenentry;
GtkWidget *rgbload_blueentry;
char redfile[1024];
char greenfile[1024];
char bluefile[1024];


void copyfile (image_type* src, image_type* dest)
{
char* p; long idx;
float fmin, fmax;
int x,y,z;

	if (allocate_image (dest,B_CHAR, src->xmax,src->ymax,src->zmax)==-1)
		fatal_memerr();
		
	p = dest->data; idx=0;
	if (src->imgtype==B_FLOAT)
	{
		fmin = src->minval.fmin; fmax = src->maxval.fmax;
	}
	else
	{
		fmin = src->minval.imin; fmax = src->maxval.imax;
	}
		
	for (z=0; z<src->zmax; z++)
		for (y=0; y<src->ymax; y++)
			for (x=0; x<src->xmax; x++)
				p[idx++] = (int) (255*(readbuf_flt (src,x,y,z)-fmin)/(fmax-fmin));
}


void execute_rgb_load ()
{
image_type r,g,b;
int x,y,z,x1,y1,z1;
unsigned char* p1; unsigned char* p2; long idx;
int esize, cnt, warnflag;
char s[255];

	r.data=NULL; b.data=NULL; g.data=NULL;
	x1=0; y1=0; z1=0; cnt=0; esize=3; warnflag=0;
	
	if (redfile[0])
	{
		file_read (redfile);	/* Reads into mainimg */
		copyfile (&mainimg, &r);
		close_file ();	/* Discard old mainimg */
		if (r.xmax>x1) x1=r.xmax;
		if (r.ymax>y1) y1=r.ymax;
		if (r.zmax>z1) z1=r.zmax;
	}
	if (greenfile[0])
	{
		file_read (greenfile);	/* Reads into mainimg */
		copyfile (&mainimg, &g);
		close_file ();	/* Discard old mainimg */
		if (g.xmax>x1) x1=g.xmax;
		if (g.ymax>y1) y1=g.ymax;
		if (g.zmax>z1) z1=g.zmax;
	}
	if (bluefile[0])
	{
		file_read (bluefile);	/* Reads into mainimg */
		copyfile (&mainimg, &b);
		close_file ();	/* Discard old mainimg */
		if (b.xmax>x1) x1=b.xmax;
		if (b.ymax>y1) y1=b.ymax;
		if (b.zmax>z1) z1=b.zmax;
	}
	
	/* Finally, we allocate the main image as B_RGB and copy all three channels */
	
	if (allocate_image (&mainimg,B_RGB, x1,y1,z1)==-1)
		fatal_memerr();
	
	if (r.data)
	{
		p1=mainimg.data; p2=r.data; idx=0;
		for (z=0; z<r.zmax; z++)
			for (y=0; y<r.ymax; y++)
				for (x=0; x<r.xmax; x++)
					p1[3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = p2[idx++];
		if ((mainimg.xmax!=r.xmax) || (mainimg.ymax!=r.ymax) || (mainimg.zmax!=r.zmax)) warnflag=1;
		freebuf (&r); cnt++;
	}
	if (g.data)
	{
		p1=mainimg.data; p2=g.data; idx=0;
		for (z=0; z<g.zmax; z++)
			for (y=0; y<g.ymax; y++)
				for (x=0; x<g.xmax; x++)
					p1[1+3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = p2[idx++];
		if ((mainimg.xmax!=g.xmax) || (mainimg.ymax!=g.ymax) || (mainimg.zmax!=g.zmax)) warnflag=1;
		freebuf (&g); cnt++;
	}
	if (b.data)
	{
		p1=mainimg.data; p2=b.data; idx=0;
		for (z=0; z<b.zmax; z++)
			for (y=0; y<b.ymax; y++)
				for (x=0; x<b.xmax; x++)
					p1[2+3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = p2[idx++];
		if ((mainimg.xmax!=b.xmax) || (mainimg.ymax!=b.ymax) || (mainimg.zmax!=b.zmax)) warnflag=1;
		freebuf (&b); cnt++;
	}

	/* No way around a sweep over the data - we need to determine min and max */
	make_minmax (&mainimg);
		
	/* Display image */
	prepare_display_window (infile);
	/* Finally, create an imlib image for display */
	make_imlib_buffer_from_main_image (0);
	file_ok = 1;
	redraw (0);

	sprintf (s, "%d files imported. Size = %li MB", 
			cnt, mainimg.xmax*mainimg.ymax*mainimg.zmax*esize/1024576);
	statusbar_display (s);
	if (warnflag)
		ShowMessage ("Warning: Not all images had same size");
	

}


void on_rgbload_bluebrowse_clicked (GtkButton *button, gpointer user_data)
{
gchar* selectedfile;

	selectedfile = ChooseFile ("Select File to load into BLUE channel");
	if (selectedfile)
		strcpy (bluefile, (char*)selectedfile);
	else
		bluefile[0]=0;
	gtk_entry_set_text (GTK_ENTRY(rgbload_blueentry), bluefile);
}


void on_rgbload_greenbrowse_clicked (GtkButton *button, gpointer user_data)
{
gchar* selectedfile;

	selectedfile = ChooseFile ("Select File to load into GREEN channel");
	if (selectedfile)
		strcpy (greenfile, (char*)selectedfile);
	else
		greenfile[0]=0;
	gtk_entry_set_text (GTK_ENTRY(rgbload_greenentry), greenfile);

}


void on_rgbload_redbrowse_clicked (GtkButton *button, gpointer user_data)
{
gchar* selectedfile;

	selectedfile = ChooseFile ("Select File to load into RED channel");
	if (selectedfile)
		strcpy (redfile, (char*)selectedfile);
	else
		redfile[0]=0;
	gtk_entry_set_text (GTK_ENTRY(rgbload_redentry), redfile);
}


void on_rgbload_extrabutton_clicked (GtkButton *button, gpointer user_data)
{

}


void on_rgbload_cancelbutton_clicked (GtkButton *button, gpointer user_data)
{
   gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}


void on_rgbload_okbutton_clicked (GtkButton *button, gpointer user_data)
{
	strcpy (redfile,gtk_entry_get_text (GTK_ENTRY(rgbload_redentry)));
	strcpy (greenfile,gtk_entry_get_text (GTK_ENTRY(rgbload_greenentry)));
	strcpy (bluefile,gtk_entry_get_text (GTK_ENTRY(rgbload_blueentry)));

	if (redfile[0] || greenfile[0] || bluefile[0])	/* At least one file name given */
	{
		if (file_ok) { close_file(); file_ok=0; } /* if there is an open file, close it */
		execute_rgb_load ();
	}
	gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}



void create_rgbload_wnd (void)
{
  GtkWidget *rgbload_wnd;
  GtkWidget *rgbload_table1;
  GtkWidget *rgbload_bluebrowse;
  GtkWidget *rgbload_greenbrowse;
  GtkWidget *rgbload_redbrowse;
  GtkWidget *rgbload_label1;
  GtkWidget *rgbload_label2;
  GtkWidget *rgbload_label3;
  GtkWidget *rgbload_extrabutton;
  GtkWidget *rgbload_cancelbutton;
  GtkWidget *rgbload_okbutton;

  rgbload_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (rgbload_wnd), "rgbload_wnd", rgbload_wnd);
  gtk_window_set_title (GTK_WINDOW (rgbload_wnd), "RGB Load");
  gtk_window_set_policy (GTK_WINDOW (rgbload_wnd), TRUE, TRUE, FALSE);
  gtk_window_set_modal (GTK_WINDOW (rgbload_wnd), TRUE);

  rgbload_table1 = gtk_table_new (4, 3, FALSE);
  gtk_widget_ref (rgbload_table1);
  gtk_object_set_data_full (GTK_OBJECT (rgbload_wnd), "rgbload_table1", rgbload_table1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (rgbload_table1);
  gtk_container_add (GTK_CONTAINER (rgbload_wnd), rgbload_table1);


  rgbload_label1 = MakeLabel (rgbload_wnd, rgbload_table1, "File for Red:",
  						"rgbload_label1", 0,0);
  rgbload_label2 = MakeLabel (rgbload_wnd, rgbload_table1, "File for Green:",
  						"rgbload_label2", 0,1);
  rgbload_label3 = MakeLabel (rgbload_wnd, rgbload_table1, "File for Blue:",
  						"rgbload_label3", 0,2);

  rgbload_redbrowse = Button_in_table (rgbload_wnd, rgbload_table1, "Browse...",
  						"rgbload_redbrowse",2,0,(gpointer)on_rgbload_redbrowse_clicked);
  rgbload_greenbrowse = Button_in_table (rgbload_wnd, rgbload_table1, "Browse...",
  						"rgbload_greenbrowse",2,1,(gpointer)on_rgbload_greenbrowse_clicked);
  rgbload_bluebrowse = Button_in_table (rgbload_wnd, rgbload_table1, "Browse...",
  						"rgbload_bluebrowse",2,2,(gpointer)on_rgbload_bluebrowse_clicked);
  rgbload_okbutton =  Button_in_table (rgbload_wnd, rgbload_table1, "OK",
  						"rgbload_okbutton",0,3,(gpointer)on_rgbload_okbutton_clicked);
  rgbload_cancelbutton =  Button_in_table (rgbload_wnd, rgbload_table1, "Cancel",
  						"rgbload_cancelbutton",1,3,(gpointer)on_rgbload_cancelbutton_clicked);
  rgbload_extrabutton  =  Button_in_table (rgbload_wnd, rgbload_table1, "No Op",
  						"rgbload_extrabutton",2,3,(gpointer)NULL);

  
  rgbload_redentry = gtk_entry_new ();
  gtk_widget_ref (rgbload_redentry);
  gtk_object_set_data_full (GTK_OBJECT (rgbload_wnd), "rgbload_redentry", rgbload_redentry,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (rgbload_redentry);
  gtk_table_attach (GTK_TABLE (rgbload_table1), rgbload_redentry, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND), 5, 3);

  rgbload_greenentry = gtk_entry_new ();
  gtk_widget_ref (rgbload_greenentry);
  gtk_object_set_data_full (GTK_OBJECT (rgbload_wnd), "rgbload_greenentry", rgbload_greenentry,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (rgbload_greenentry);
  gtk_table_attach (GTK_TABLE (rgbload_table1), rgbload_greenentry, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND), 5, 3);

  rgbload_blueentry = gtk_entry_new ();
  gtk_widget_ref (rgbload_blueentry);
  gtk_object_set_data_full (GTK_OBJECT (rgbload_wnd), "rgbload_blueentry", rgbload_blueentry,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (rgbload_blueentry);
  gtk_table_attach (GTK_TABLE (rgbload_table1), rgbload_blueentry, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND), 5, 3);

  gtk_signal_connect (GTK_OBJECT (rgbload_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_rgbload_cancelbutton_clicked),
                      rgbload_wnd);


	gtk_widget_show (rgbload_wnd);
}


void rgbload_main ()
{
	create_rgbload_wnd();

}


/**************************************************************

Import multiple: Use standard file dialog, but interpret
the result. If it is a directory, select all. If it is a mask,
do a regexp comparison (???)

****************************************************************/

/* Regexp interpreter - copied from the zip source.
 Credits:
 Mark Adler, Richard B. Wales, Jean-loup Gailly,
 Kai Uwe Rommel, Onno van der Linden and Igor Mandrichenko.
*/
 

#define case_map(c) (c)

/* If p is a sh expression, a pointer to the first special character is
   returned.  Otherwise, NULL is returned. */


char *isshexp(char* p)
{

	for (; *p; p++)
		if (*p == '\\' && *(p+1))
			p++;
		else if (*p == '?' || *p == '*' || *p == '[')
			return p;
	return NULL;
}


/* Recursively compare the sh pattern p with the string s and return 1 if
   they match, and 0 or 2 if they don't or if there is a syntax error in the
   pattern.  This routine recurses on itself no deeper than the number of
   characters in the pattern.
   p is the regexp pattern
   s is the string which does or does not match p
*/

int recmatch(unsigned char* p, unsigned char* s)
{
unsigned int c;       /* pattern char or start of range in [-] loop */

	/* Get first character, the pattern for new recmatch calls follows */
	c = *p++;

	/* If that was the end of the pattern, match if string empty too */
	if (c == 0) return *s == 0;

	/* '?' (or '%') matches any character (but not an empty string) */
	if (c == '?')
		return *s ? recmatch(p, s + 1) : 0;

	/* '*' matches any number of characters, including zero */
	if (c == '*')
	{
		if (*p == 0) return 1;
		for (; *s; s++)
			if ((c = recmatch(p, s)) != 0)
				return (int)c;
		return 2;           /* 2 means give up--shmatch will return false */
	}

	/* Parse and process the list of characters and ranges in brackets */
	if (c == '[')
	{
		int e;              /* flag true if next char to be taken literally */
		unsigned char *q;   /* pointer to end of [-] group */
		int r;              /* flag true to match anything but the range */

		if (*s == 0)                        /* need a character to match */
			return 0;

		p += (r = (*p == '!' || *p == '^')); /* see if reverse */
		for (q = p, e = 0; *q; q++)         /* find closing bracket */
			if (e)
				e = 0;
			else
				if (*q == '\\')
					e = 1;
				else if (*q == ']')
					break;

		if (*q != ']')                      /* nothing matches if bad syntax */
			return 0;

		for (c = 0, e = *p == '-'; p < q; p++)      /* go through the list */
		{
			if (e == 0 && *p == '\\')         /* set escape flag if \ */
				e = 1;
			else if (e == 0 && *p == '-')     /* set start of range if - */
				c = *(p-1);
			else
			{
				unsigned char cc = case_map(*s);
				if (*(p+1) != '-')
					for (c = c ? c : *p; c <= *p; c++)    /* compare range */
						if (case_map(c) == cc)
							return r ? 0 : recmatch(q + 1, s + 1);

				c = e = 0;                      /* clear range, escape flags */
			}
		}
		return r ? recmatch(q + 1, s + 1) : 0;      /* bracket match failed */
	}

	/* If escape ('\'), just compare next character */
	if (c == '\\')
		if ((c = *p++) == 0)                /* if \ at end, then syntax error */
			return 0;

	/* Just a character--compare it */
	return case_map(c) == case_map(*s) ? recmatch(p, ++s) : 0;
}


/* Compare the sh pattern p with the string s and return true if they match,
   false if they don't or if there is a syntax error in the pattern.
	*p   sh pattern to match 
	*s   string to match it to 
*/


int shmatch(char* p, char* s)
{
	return recmatch((unsigned char *) p, (unsigned char *) s) == 1;
}




void strswap (char* p1, char* p2)
{
int i;
char x;

	for (i=0; i<1024; i++)
	{
		x=p1[i]; p1[i]=p2[i]; p2[i]=x;
	}
}



int internal_open_multiple (gchar *selectedfile, image_type *auximg)
{
gchar dir[1024]; gchar filemask[256]; gchar resfile[2048]; char s[2048];
int e,i,j,k,err;
struct stat statresult;		/* to check if a file is a directory */
DIR* directory;
struct dirent* entry;
int filecnt, zbuf;
long framelen;
image_type framebuf;
char* p1; char* p2;
char* filelist;
int minidx;
char tiffreport_save;


	err = stat (selectedfile, &statresult);
	if ((!err) && (S_ISDIR (statresult.st_mode)))		/* Does the file name specify a directory? */
	{
		if (selectedfile[strlen(selectedfile)-1]!='/')
			strcat (selectedfile, "/");	/* all in this directory */
		strcat (selectedfile, "*");	/* all in this directory */
	}
	dp (1, "File(s) selected: %s\n", selectedfile);
	
	/* Now split selectedfile into directory and file spec. TODO - isn't there a library function?
			see "basename" and "dirname" functions */
	
	p1 = strrchr (selectedfile, '/');
	if (!p1)										/* If NULL, no '/' found in name */
	{
		strcpy (filemask, selectedfile);			/* in that case, use all for file name */
		strcpy (dir, ".");							/* and . (dot) for current directory */
	}
	else
	{
		strcpy (filemask, p1+1);					/* file name starts right after '/' */
		p2 = selectedfile;
		k=0;
		while (p2<p1) dir[k++] = *p2++;				/* Copy directory name up to '/' */
		dir[k] = 0;									/* Null-terminate */
	}
	dp (2, "Split: %s --- %s\n",dir,filemask);
	
	/* Do a directory scan, pass 1: determine number of files */
	
	filecnt=0;
	directory = opendir (dir);
	do
	{
		entry = readdir (directory);
		if (entry) 
		{
			strcpy (resfile, dir); strcat (resfile, "/");
			strcat (resfile, entry->d_name);
			err=stat (resfile, &statresult);
			if ( (!S_ISDIR (statresult.st_mode))	/* Skip directories */
				&& (!err)							/* error STATing */
				&& (shmatch (filemask, entry->d_name)))
			{
				filecnt ++;
			}
		}
	}
	while (entry);
	
	/* Pass 2: Now that we know the number of files, allocate
				memory for the filename list and sort the list.
				This could be done way more elegantly using a linked 
				list, particularly since we have GTK. But at this time,
				I don't have my GTK docu here, so leave it for later */

	filelist = (char*) calloc (filecnt, 1024);
	if (!filelist) fatal_memerr();
	
	rewinddir (directory); i=0;
	do
	{
		entry = readdir (directory);
		if (entry) 
		{
			strcpy (resfile, dir); strcat (resfile, "/");
			strcat (resfile, entry->d_name);
			err=stat (resfile, &statresult);
			if ( (!S_ISDIR (statresult.st_mode))	/* Skip directories */
				&& (!err)							/* error STATing */
				&& (shmatch (filemask, entry->d_name)))
			{
				strcpy (&(filelist[i*1024]), entry->d_name);
				i++;
			}
		}	
	}
	while (entry);

	/* Dirtysort algorithm. There are better ways, but... */
	
	if (filecnt>1)
	{
		for (i=0; i<filecnt-1; i++)
		{
			minidx=i;
			for (j=i+1; j<filecnt; j++)
			{
				if (strcmp(&(filelist[j*1024]),&(filelist[minidx*1024]))<0)
				{
					minidx=j;
				}
			}
			if (minidx != i) strswap (&(filelist[i*1024]),&(filelist[minidx*1024]));
		}
	}
	

	dp (1, "Reading %d files from %s\n",filecnt,selectedfile);

	/* Pass 3: Open the first file to determine size and type
				then allocate memory
				then read the remaining files.
				Abort if any size/type doesn't match */
				
	tiffreport_save = tiffreport; tiffreport=0;
	
	for (i=0; i<filecnt; i++)
	{
		strcpy (resfile, dir); strcat (resfile, "/");
		strcat (resfile, &(filelist[i*1024]));
		dp (1, "Opening %s\n", resfile);

		internal_file_read (resfile, &framebuf);
/*			if (mainimg.zmax>1)
		{
			sprintf (s, "Error: can't open stacks!\nFile; %s",resfile);
			ShowMessageC (s,"Multi-Open failed");
			i=-1;
		}*/
		if (i==0)	/* First file! */
		{
			if (allocate_image (auximg, framebuf.imgtype, framebuf.xmax, framebuf.ymax, filecnt*framebuf.zmax)==-1)
			{
				fatal_memerr ();
				i=-1;
			}
			zbuf = framebuf.zmax;		/* Buffer stack size for later size check */
			dp (2, "Multi-Open: Allocated image (%d,%d,%d) type %d\n", auximg->xmax, auximg->ymax, auximg->zmax, auximg->imgtype);
		}
		/* Check for matching size and type */
		if ((auximg->xmax!=framebuf.xmax) || (auximg->ymax!=framebuf.ymax) ||  (framebuf.zmax!=zbuf) || (auximg->imgtype!=framebuf.imgtype))
		{
			sprintf (s, "Error: Size/type does not match!\nFile; %s",resfile);
			ShowMessageC (s,"Multi-Open failed");
			i=-1;
		}
		if (i>=0)	/* Copy one slice into auximg */
		{
			p1=framebuf.data; p2=auximg->data;
			e=DATALEN(framebuf.imgtype); 
			framelen = (long)framebuf.xmax*(long)framebuf.ymax*(long)zbuf*e;
			p2 += i*framelen;  /* Advance to correct z-position */
			memcpy (p2,p1, framelen);
		}
		dp (1, "file %d done: %s\n",i+1,resfile);
				
		/* Discard mainimg memory */
		freebuf (&framebuf);
	}
	
	closedir (directory);
	tiffreport = tiffreport_save;
	free (filelist);
	
	/* If all went well, copy auximg into mainimg and prepare display */
	
	if (i>=0)
		return 0;

	return 1;


}


/* Main "open multiple" with file select gui and preparation of mainimage */

void do_open_multiple ()
{
gchar* selectedfile;
int err;
image_type auximg;


	selectedfile = ChooseFile("Select file(s) to load:");

	if (selectedfile)
	{
		err = internal_open_multiple (selectedfile, &auximg);
		
		if (!err)
		{
			close_file ();						/* Clean up the previous image */
			mainimg.data = auximg.data;
			mainimg.xmax=auximg.xmax; mainimg.ymax=auximg.ymax; mainimg.zmax=auximg.zmax;
			mainimg.imgtype = auximg.imgtype; mainimg.uid = auximg.uid;
			/* No way around a sweep over the data - we need to determine min and max */
			make_minmax (&mainimg);
			
			/* Display image */
			prepare_display_window (infile);
			/* Finally, create an imlib image for display */
			make_imlib_buffer_from_main_image (0);
			file_ok = 1;
			redraw (0);

		}	
	}

}

		

/* we used this further up - since it's a lot of typing, we save it here:

	if (err==-1)
	{
		switch (errno)
		{
			case EBADF: fprintf (stderr, "filedes is bad.\n"); break;
			case ENOENT: fprintf (stderr, "A component of the path file_name does  not  exist,\n"); break;
			case ENOTDIR: fprintf (stderr, "A component of the path is not a directory.\n"); break;
			case ELOOP: fprintf (stderr, "Too  many symbolic links encountered\n"); break;
			case EACCES: fprintf (stderr, "Permission denied.\n"); break;
			case ENOMEM: fprintf (stderr,"Out of memory (i.e. kernel memory)\n"); break;
			case ENAMETOOLONG: fprintf (stderr, "File name too long.\n"); break;
		}
	}

*/

