
/*
 * Misc Op: Miscellaneous operators
 */
 

/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */

 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "miscop.h"
#include "roi.h"
#include "viz.h"


/*************************************************************

	Logexp: Compute log(image), log10(imge), exp(image), 
	10^image, x^image and image^x (mainly for macro functions)
	TODO: The monadic macro functions should use these
			for consistency.
	TODO: Sucks with RGB, a common malady.

**************************************************************/


void logexp (image_type* img, int what, float para)
{
image_type buf;
float *p;
int x,y,z; long idx;

	if (allocate_image (&buf, B_FLOAT, img->xmax,img->ymax,img->zmax)==-1)
		fatal_memerr();
	p = buf.data;
		
	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				idx = x + img->xmax*(y + img->ymax*z);
				switch (what)
				{
					case 0:
						p[idx] = SAFELOG ( readbuf_flt(img,x,y,z));
						break;
					case 1:
						p[idx] = SAFELOG10 ( readbuf_flt(img,x,y,z));
						break;
					case 2:
						p[idx] = exp ( readbuf_flt(img,x,y,z));
						break;
					case 3:
						p[idx] = pow ( 10.0, readbuf_flt(img,x,y,z));
						break;
					case 4:
						p[idx] = pow ( readbuf_flt(img,x,y,z), para);
						break;
					case 5:
						p[idx] = pow ( para, readbuf_flt(img,x,y,z));
						break;
				}
			}
			
	handback (img, &buf);

}



/*************************************************************

	Center: Center features of an image (i.e. shift the
	centroid in the middle of the image area)
	Depending on mass, we use nonzero pixels, equally weighted
	or we weight them by gray value (mass=1)

**************************************************************/


vertex get_centroid (image_type* img, int slice, int mass)
{
int x,y;
float cogx, cogy, val, buf;
vertex v;

	val=0.0; cogx=0.0; cogy=0.0;
	
	if (mass)
	{
	
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,slice);
				val+=buf;
				cogx+=buf*x;
				cogy+=buf*y;
			}
		if (val>0)
		{
			cogx/=val;
			cogy/=val;
		}
		else
		{
			cogx=img->xmax/2;
			cogy=img->ymax/2;
		}
	}
	else
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,slice);
				if (buf!=0.0)
				{
					val+=1.0;
					cogx+=x;
					cogy+=y;
				}
			}
		if (val>0)
		{
			cogx/=val;
			cogy/=val;
		}
		else
		{
			cogx=img->xmax/2;
			cogy=img->ymax/2;
		}
	}
	v.x = (int)(0.5+cogx);
	v.y = (int)(0.5+cogy);
	return v;
}			




void center_feat (image_type* img, int z, int mass)
{
vertex v;
int x,y,x1,y1,dx,dy;
image_type temp;

	if (allocate_image (&temp, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();


	v = get_centroid (img, z, mass);
	dp (1, "Center image: Centroid at %d,%d\n",v.x,v.y);

	dx = v.x - img->xmax/2; /* Negative: Shift right */
	dy = v.y - img->ymax/2;
	if ((dx==0) &&(dy==0)) return;


	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char c;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; c = 0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &c, x1,y1,z);
					writebuf (&temp, &c, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_SHORT:
		{
			short s;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; s = 0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &s, x1,y1,z);
					writebuf (&temp, &s, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_FLOAT:
		{
			float f;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; f = 0.0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &f, x1,y1,z);
					writebuf (&temp, &f, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_RGB:
		{
			rgbtriplet rgb;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; rgb.red=0; rgb.green=0; rgb.blue=0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &rgb, x1,y1,z);
					writebuf (&temp, &rgb, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
	}

	freebuf (&temp);

}







/*************************************************************

	Rotate: Rotate an image by theta degrees
	
**************************************************************/

/* Original and rotated image have the same dimensions.
   whatever does not fit will be cropped. Undefined areas
   are set to zero. 
   THETA IS GIVEN IN DEGREES! 
   
   If rotation angles are between 45 and 135� or between 
   225 and 315 degrees, we flip image's X and Y extensions
   
*/


void rotate_img (image_type* img, double theta)
{
double st, ct;	/* cosine and sine of theta */
int x,y,z;		/* target image coords */
double x1,y1; 		/* Original image coords */
double xh,yh;	/* Half-image coordinates */
float* p;
rgbtriplet* prgb;
long i, idx, zofs;
int new_xmax, new_ymax;


	p = (float*) calloc (img->xmax * img->ymax, sizeof (float));
	if (!p) fatal_memerr();
	prgb = (rgbtriplet*)p; /* !!! This assumes that sizeof (float) > sizeof (rgbtriplet) ...or SIGSEGV */

	new_xmax = img->xmax; 
	new_ymax = img->ymax; 
	if ( ((fabs(theta)>45) && (fabs(theta)<135)) || ((fabs(theta)>225) && (fabs(theta)<315)) ) 
	{
		new_xmax = img->ymax;
		new_ymax = img->xmax;
	}

	st = sin(PI*theta/180.0);
	ct = cos(PI*theta/180.0);
	xh = new_xmax/2.0;
	yh = new_ymax/2.0;

	for (z=0; z<img->zmax; z++)		/* slice by slice */
	{
		zofs = new_xmax * new_ymax * z;
		dp (1, "Rotating slice %d at offset %li\n", z, zofs);
		for (y=0; y < new_ymax; y++)
		{
			idx = y*new_xmax;
			for (x=0; x < new_xmax; x++)
			{
				x1 = (x-xh)*ct + (y-yh)*st + img->xmax/2.0;
				y1 = -(x-xh)*st + (y-yh)*ct + img->ymax/2.0;
				if ((y1<0) || (x1<0) || (x1>=img->xmax) || (y1>=img->ymax))
				{
					if (img->imgtype==B_RGB)
						prgb[idx+x]=(rgbtriplet){0,0,0};
					else
						p[idx+x]=0;
				}
				else
				{
					if (img->imgtype==B_RGB)
						prgb [idx+x] = rgbireadbuf (img, x1,y1,z, intp_pref);
					else
						p [idx+x] = ireadbuf (img, x1,y1,z, intp_pref);
				}
			}
		}
		
		/* copy the slice back into the original image */
		
		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p1; float buf;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
				{
					buf = IROUND(p[i]); 
					if (buf<0) buf=0; else if (buf>255) buf=255;
					p1[i+zofs] = (int)buf;
				}
				break;
			}
			case B_SHORT:
			{
				short* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = IROUND(p[i]);
				break;
			}
			case B_FLOAT:	/* Beauty! We don't have to round here!!! */
			{
				float* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = p[i];
				break;
			}						
			case B_RGB:	
			{
				rgbtriplet* p1;
				p1 = img->data;
				for (i=0; i < new_xmax * new_ymax; i++)
					p1[i+zofs] = prgb[i];
				break;
			}
		}
	}
	
	img->xmax = new_xmax; img->ymax = new_ymax;
	img->uid = next_uid++;
	free (p);
}


/* Rescale image, scale by xfac, yfac. */


void zoom2D (image_type* img, int xnew, int ynew)
{
image_type tempimg;
int x,y,z;
float dx,dy,fy,fx;
long idx;



	if (allocate_image (&tempimg, img->imgtype, xnew,ynew,img->zmax)==-1)
		fatal_memerr();
	
	/* interpolation step size */
	
	dx = (float)img->xmax / (float)xnew;
	dy = (float)img->ymax / (float)ynew;
	
	/* Now transfer. Go via float, because we have to interpolate anyway */
	
	idx=0;
	for (z=0; z<tempimg.zmax; z++)
	{
		for (y=0, fy=0; y<tempimg.ymax; y++, fy+=dy)
			for (x=0, fx=0; x<tempimg.xmax; x++, fx+=dx)
			{
				switch (tempimg.imgtype)
				{
					case B_CHAR:
					{
						unsigned char *p = tempimg.data;
						p[idx] = (int) (0.5+ireadbuf (img, fx,fy,z,intp_pref));
						break;
					}
					case B_SHORT:
					{
						short *p = tempimg.data;
						p[idx] = (int) (0.5+ireadbuf (img, fx,fy,z,intp_pref));
						break;
					}
					case B_FLOAT:
					{
						float *p = tempimg.data;
						p[idx] = ireadbuf (img, fx,fy,z,intp_pref);
						break;
					}
					case B_RGB:
					{
						rgbtriplet* prgb = tempimg.data;
						prgb [idx] = rgbireadbuf (img, fx,fy,z, intp_pref);
						break;
					}
				}
				
				idx++;
			}
	}

	handback (img, &tempimg);

}






/* Interpolate a number of Z slices into existing ones. 
   numslices is the number of slices to create between two existing ones */

/* TODO: Cannot yet handle RGB. Remember that we have a function "rgbireadbuf" !!! */

void interpol_slice (image_type* img, int numslices)
{
image_type target;
image_type oneslice;
int znew,c;
int x,y,z; long idx;
float buf;

	if (img->imgtype == B_RGB) return;
	if (img->zmax==1) return;

	znew = (img->zmax-1)*(numslices+1)+1;	/* New z dimensions */
	
	/* Allocate the intermediate and final image */
	
	if (allocate_image (&oneslice,B_FLOAT,img->xmax,img->zmax,1)==-1)
		fatal_memerr();

	if (allocate_image (&target,img->imgtype,img->xmax,img->ymax,znew)==-1)
		fatal_memerr();


	for (y=0; y<img->ymax; y++)		/* X-dominant interpolation */
	{
	
		/* Fill the x-z slice we want to interpolate from */
		
		for (x=0; x<img->xmax; x++)
			for (z=0; z<img->zmax; z++)
			{
				buf = readbuf_flt (img,x,y,z);	/* Interpolating means float, anyway */
				((float*)oneslice.data)[x+oneslice.xmax*z] = buf;
			}
		oneslice.uid = next_uid++;
		
		/* Fill and interpolate the destination */

		for (x=0; x<img->xmax; x++)
			for (z=0; z<znew; z++)
			{
				idx = x + target.xmax*(y + target.ymax*z);
				buf = 0.5*ireadbuf (&oneslice,x, (double)z/(double)(numslices+1),1,intp_pref);
				switch (target.imgtype)
				{
					case B_CHAR:
						c = (int)(0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						((unsigned char*)target.data)[idx]=c;
						break;
					case B_SHORT:
						c = (int)(0.5+buf);
						((short*)target.data)[idx]=c;
						break;
					case B_FLOAT:
						((float*)target.data)[idx]=buf;
						break;
				}
			}
	}
	
	freebuf (&oneslice);
	if (allocate_image (&oneslice,B_FLOAT,img->ymax,img->zmax,1)==-1)
		fatal_memerr();

	for (x=0; x<img->xmax; x++)		/* Y-dominant interpolation */
	{
	
		/* Fill the y-z slice we want to interpolate from */
		
		for (y=0; y<img->ymax; y++)
			for (z=0; z<img->zmax; z++)
			{
				buf = readbuf_flt (img,x,y,z);	/* Interpolating means float, anyway */
				((float*)oneslice.data)[y+oneslice.xmax*z] = buf;
			}
		oneslice.uid = next_uid++;
		
		/* Fill and interpolate the destination */

		for (y=0; y<img->ymax; y++)
			for (z=0; z<znew; z++)
			{
				idx = x + target.xmax*(y + target.ymax*z);
				buf = 0.5*ireadbuf (&oneslice,y, (double)z/(double)(numslices+1),1,intp_pref);
				switch (target.imgtype)
				{
					case B_CHAR:
						c = (int)(0.5+buf);
						if (c<0) c=0; else if (c>255) c=255;
						((unsigned char*)target.data)[idx]+=c;
						break;
					case B_SHORT:
						c = (int)(0.5+buf);
						((short*)target.data)[idx]+=c;
						break;
					case B_FLOAT:
						((float*)target.data)[idx]+=buf;
						break;
				}
			}
	}

	handback (img, &target);
	freebuf (&oneslice);

}



/* Flip the x and z coordinates (rotate around Y)
   to allow different access to the stack */


void flip_xz (image_type* img)
{
int x,y,z,e,i;
long sidx, didx;
image_type target;
unsigned char *p1, *p2;

	if (img->zmax==1) return;	/* can do this on stacks only */

	if (allocate_image (&target,img->imgtype,img->zmax,img->ymax,img->xmax)==-1)
		fatal_memerr();
	e = DATALEN(img->imgtype);
	p1 = img->data;
	p2 = target.data;

	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				sidx = e*(x + img->xmax*(y + img->ymax*z));
				didx = e*(z + target.xmax*(y + target.ymax*x));
				for (i=0; i<e; i++) p2[didx+i] = p1[sidx+i];
			}
	handback (img, &target);

}


/******************************************************

	Rot Zoom: GUI for rotate and zoom operation
	
*******************************************************/


GtkAdjustment *rotzoom_rotspin_adj;
GtkAdjustment *rotzoom_xscalespin_adj;
GtkAdjustment *rotzoom_yscalespin_adj;
GtkAdjustment *rotzoom_zscalespin_adj;
image_type* rotzoomimg;


void on_rotzoom_cancel_clicked (GtkButton *button, gpointer user_data)
{


	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_rotzoom_ok_clicked (GtkButton *button, gpointer user_data)
{
double theta;
int xnew, ynew, zadd;

	undo_nexus (&mainimg);

	theta = rotzoom_rotspin_adj->value;
	xnew = rotzoom_xscalespin_adj->value;
	ynew = rotzoom_yscalespin_adj->value;
	zadd = rotzoom_zscalespin_adj->value;

	/* Sequence of operations: Zoom first, then rotate, then add Z slices */

	if ((xnew != rotzoomimg->xmax) || (ynew != rotzoomimg->ymax))
		zoom2D (rotzoomimg, xnew,ynew);	

	if (theta !=0.0) rotate_img (rotzoomimg, theta);

	if (zadd>0)
	{
		interpol_slice (rotzoomimg, zadd);
		gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, mainimg.zmax-1);
	}

	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}





void create_rotzoom_wnd (image_type* img)
{
  GtkWidget *rotzoom_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *rotzoom_label1, *rotzoom_label2, *rotzoom_label3, *rotzoom_label4;
  GtkWidget *rotzoom_cancel;
  GtkWidget *rotzoom_ok;
  GtkWidget *rotzoom_rotspin;
  GtkWidget *rotzoom_xscalespin;
  GtkWidget *rotzoom_yscalespin;
  GtkWidget *rotzoom_zscalespin;
  GtkWidget* hsep1;
  GtkWidget* hsep2;

	rotzoomimg = img;

	rotzoom_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (rotzoom_wnd), "rotzoom_wnd", rotzoom_wnd);
	gtk_window_set_title (GTK_WINDOW (rotzoom_wnd), "Rotate - Zoom");
	gtk_window_set_modal (GTK_WINDOW (rotzoom_wnd), TRUE);

	vbox1 = vbox_in_window (rotzoom_wnd, "rotzoom_vbox1");
	table1 = MakeTable (rotzoom_wnd, vbox1, "table1", 2,7);
	rotzoom_label1 = MakeLabel (rotzoom_wnd,table1,"Rotate (degrees)","rotzoom_label1",0,0);
	rotzoom_label2 = MakeLabel (rotzoom_wnd,table1,"New X dimension","rotzoom_label2",0,1);
	rotzoom_label3 = MakeLabel (rotzoom_wnd,table1,"New Y dimension","rotzoom_label3",0,2);
	rotzoom_label4 = MakeLabel (rotzoom_wnd,table1,"Add new Z slices","rotzoom_label4",0,3);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label1), 0, 0.5);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label2), 0, 0.5);
	gtk_misc_set_alignment (GTK_MISC (rotzoom_label3), 0, 0.5);

	rotzoom_rotspin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, -359, 359, 1, 10, 0));
	rotzoom_rotspin = MakeSpinbutton (rotzoom_rotspin_adj,rotzoom_wnd,table1,"rotzoom_rotspin",1,0);

	rotzoom_xscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->xmax, 16, 32768, 1, 10, 0));
	rotzoom_xscalespin = MakeSpinbutton (rotzoom_xscalespin_adj,rotzoom_wnd,table1,"rotzoom_xscalespin",1,1);

	rotzoom_yscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (img->ymax, 16, 32768, 1, 10, 0));
	rotzoom_yscalespin = MakeSpinbutton (rotzoom_yscalespin_adj,rotzoom_wnd,table1,"rotzoom_yscalespin",1,2);

	rotzoom_zscalespin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, 10, 1, 10, 0));
	rotzoom_zscalespin = MakeSpinbutton (rotzoom_zscalespin_adj,rotzoom_wnd,table1,"rotzoom_zscalespin",1,3);

	hsep1 = MakeHsep (rotzoom_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (rotzoom_wnd,table1,"hsep1",1,4);

	rotzoom_cancel = Button_in_table (rotzoom_wnd,table1,"Cancel","rotzoom_cancel",1,5,NULL);
	rotzoom_ok = Button_in_table (rotzoom_wnd,table1,"OK","rotzoom_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (rotzoom_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_rotzoom_cancel_clicked), rotzoom_wnd);
	gtk_signal_connect (GTK_OBJECT (rotzoom_ok), "clicked",
                    	GTK_SIGNAL_FUNC (on_rotzoom_ok_clicked), rotzoom_wnd);
	gtk_signal_connect (GTK_OBJECT (rotzoom_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_rotzoom_cancel_clicked), rotzoom_wnd);

	gtk_widget_show (rotzoom_wnd);
}


void rotzoom_main ()
{

	create_rotzoom_wnd (&mainimg);
	
}



/*************************************************************

	Rebinning function. Has a similar effect as zoom,
	but zooms by integer factors through binning
	Rebinning has the advantage of noise reduction
	so if you need to reduce by a factor of e.g., 2, rebin.

**************************************************************/

/* Note: Rebinning will take place without respect of size,
	Image values are assumed to continue constant beyond image boundaries
*/




void rebin_img (image_type *img, int rebin_ngh)
{
int x,x1,y,y1,z,xm,ym,x2,y2,x3,y3,c;
image_type target;
float buf,r,g,b; int cnt;
rgbtriplet rgb;

	/* First calculate target image size and allocate it */

	if (rebin_ngh<=1) return;		/* Should not happen */

	xm = img->xmax / rebin_ngh;		/* Truncating division is intentional */
	ym = img->ymax / rebin_ngh;

	if (allocate_image (&target, img->imgtype, xm,ym,img->zmax)==-1)
		fatal_memerr();

	for (z=0; z<img->zmax; z++)
	{
		x3=0; y3=0;
		for (y=0; y<img->ymax; y+=rebin_ngh)
		{
			x3=0;
			for (x=0; x<img->xmax; x+=rebin_ngh)
			{

				buf= r = g = b = 0.0; cnt=0;
				for (y1=0; y1<rebin_ngh; y1++)
					for (x1=0; x1<rebin_ngh; x1++)
					{
						x2 = MIN(x+x1, img->xmax-1);
						y2 = MIN(y+y1, img->ymax-1);
						if (img->imgtype==B_RGB)
						{
							rgb = readbuf_rgb (img,x2,y2,z);
							r+=rgb.red; b+=rgb.blue; g+=rgb.green; cnt++;
						}
						else
						{
							buf += readbuf_flt (img,x2,y2,z);
							cnt++;
						}
					}

				if (img->imgtype==B_RGB)
				{
					r/=cnt; b/=cnt; g/=cnt;
				}
				else
					buf /= cnt; 

				if ((x3<xm) && (y3<ym))
				{
					switch (img->imgtype)
					{
						case B_CHAR:
							c = IROUND(buf); if (c<0) c=0; else if (c>255) c=255;
							((unsigned char *)target.data)[x3+xm*(y3+ym*z)] = c;
							break;
						case B_SHORT:
							c = IROUND(buf);
							((short *)target.data)[x3+xm*(y3+ym*z)] = c;
							break;
						case B_FLOAT:
							((float *)target.data)[x3+xm*(y3+ym*z)] = buf;
							break;
						case B_RGB:
							c = IROUND(r); if (c<0) c=0; else if (c>255) c=255;
							rgb.red=c;
							c = IROUND(g); if (c<0) c=0; else if (c>255) c=255;
							rgb.green=c;
							c = IROUND(b); if (c<0) c=0; else if (c>255) c=255;
							rgb.blue=c;
							((rgbtriplet *)target.data)[x3+xm*(y3+ym*z)] = rgb;
							break;
					}
				}

				x3++;
			}
			y3++;
		}
	}
	make_minmax (&target);
	handback (img, &target);

}



int rebin_func;


void on_rebin_cancel_clicked (GtkButton *widget, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_rebin_ok_clicked (GtkButton *widget, gpointer user_data)
{
	undo_nexus (&mainimg);
	rebin_img (&mainimg, rebin_func);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}



void rebin_menu_handler (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	rebin_func = i;
}



void create_rebin_wnd (image_type* img)
{
  GtkWidget *rebin_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *rebin_label1;
  GtkWidget *rebin_cancel;
  GtkWidget *rebin_ok;
  GtkWidget *menuitem;
  GtkWidget *rebin_menu;
  GtkWidget *rebin_menu_menu;
  GtkWidget* hsep1;
  GtkWidget* hsep2;

	rebin_func=2;

	rebin_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (rebin_wnd), "rebin_wnd", rebin_wnd);
	gtk_window_set_title (GTK_WINDOW (rebin_wnd), "Image Binning");
	gtk_window_set_modal (GTK_WINDOW (rebin_wnd), TRUE);

	vbox1 = vbox_in_window (rebin_wnd, "rebin_vbox1");
	table1 = MakeTable (rebin_wnd, vbox1, "table1", 2,1);
	rebin_label1 = MakeLabel (rebin_wnd,table1,"Rebinning group","rebin_label1",0,0);
	rebin_menu = Make_Optionmenu (rebin_wnd,table1,"rebin_opt",1,0);
	rebin_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "2 x 2", (gpointer) &menuopt2);
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "3 x 3", (gpointer) &menuopt3);
	menuitem = Option_menuitem (rebin_menu_menu, GTK_SIGNAL_FUNC(rebin_menu_handler), "4 x 4", (gpointer) &menuopt4);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (rebin_menu), rebin_menu_menu);


	hsep1 = MakeHsep (rebin_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (rebin_wnd,table1,"hsep1",1,4);

	rebin_cancel = Button_in_table (rebin_wnd,table1,"Cancel","rebin_cancel",1,5,NULL);
	rebin_ok = Button_in_table (rebin_wnd,table1,"OK","rebin_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (rebin_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_rebin_cancel_clicked), rebin_wnd);
	gtk_signal_connect (GTK_OBJECT (rebin_ok), "clicked",
                    	GTK_SIGNAL_FUNC (on_rebin_ok_clicked), rebin_wnd);
	gtk_signal_connect (GTK_OBJECT (rebin_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_rebin_cancel_clicked), rebin_wnd);

	gtk_widget_show (rebin_wnd);
}


void rebin_main ()
{

	create_rebin_wnd (&mainimg);
	
}



/************************************************************+

	Edit the clickstack
	
**************************************************************/


typedef struct
{
	GtkAdjustment *xadj;	/* The adjustment part of a spin button */
	GtkWidget* xspin;		/* A spin button widget */
	GtkAdjustment *yadj;	/* The adjustment part of a spin button */
	GtkWidget* yspin;		/* A spin button widget */
	GtkAdjustment *zadj;	/* The adjustment part of a spin button */
	GtkWidget* zspin;		/* A spin button widget */
	GtkWidget* label;		/* Its label text */
}
cstedit_type;				/* Used in edit_clickstack */
cstedit_type cst[CLICKSTACK_MAX];


int cst_opts[CLICKSTACK_MAX];	/* To identify the row in the "on...changed functions */




void on_cst_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_cst_x_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].x = (int)cst[i].xadj->value;
}

void on_cst_y_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].y = (int)cst[i].yadj->value;

}

void on_cst_z_adj (GtkWidget* widget, gpointer user_data)
{
int i;

	i = *(int*)user_data;
	clickstack[i].z = (int)cst[i].zadj->value;

}


void edit_clickstack (image_type* img)
{

GtkWidget *cst_wnd;
GtkWidget* vbox1;
GtkWidget *table1;
GtkWidget *l1,*l2,*l3, *l4;
GtkWidget* cst_cancelbutton;

int i,j,k;
char s[100];

	j = MIN (CLICKSTACK_MAX,clickstack_last);
	if (!j) return;	/* Nothing to edit */
	
	for (i=0; i<CLICKSTACK_MAX; i++) cst_opts[i]=i;

	cst_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (cst_wnd), "cst_wnd", cst_wnd);
	gtk_window_set_title (GTK_WINDOW (cst_wnd), "Edit Clickstack");
	gtk_window_set_modal (GTK_WINDOW (cst_wnd), TRUE);

	vbox1 = vbox_in_window (cst_wnd, "cst_vbox1");
	table1 = MakeTable (cst_wnd, vbox1, "table1", 4,j+1);

	l1 = MakeLabel (cst_wnd,table1,"Stack position","l1",0,0);
	l2 = MakeLabel (cst_wnd,table1,"X coordinate","l2",1,0);
	l3 = MakeLabel (cst_wnd,table1,"Y coordinate","l3",2,0);
	l4 = MakeLabel (cst_wnd,table1,"Z coordinate","l4",3,0);

	for (i=0; i<j; i++)
	{
		k=j-i;
		sprintf (s, "%2d", k);
		if (i==0) strcat (s, " (Bottom)");
		cst[i].label = MakeLabel (cst_wnd,table1,s,s,0,k);
		
		cst[i].xadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].x,0,img->xmax-1,1,1,0));
		cst[i].xspin= MakeSpinbutton (cst[i].xadj,cst_wnd,table1,s,1,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].xadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_x_adj), &cst_opts[i]);
		
		cst[i].yadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].y,0,img->ymax-1,1,1,0));
		cst[i].yspin= MakeSpinbutton (cst[i].yadj,cst_wnd,table1,s,2,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].yadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_y_adj), &cst_opts[i]);
		
		cst[i].zadj = GTK_ADJUSTMENT(gtk_adjustment_new (clickstack[i].z,0,img->zmax-1,1,1,0));
		cst[i].zspin= MakeSpinbutton (cst[i].zadj,cst_wnd,table1,s,3,k);
		gtk_signal_connect (GTK_OBJECT (cst[i].zadj), "value_changed",
                      GTK_SIGNAL_FUNC (on_cst_z_adj), &cst_opts[i]);
	}

	cst_cancelbutton = Button_in_table (cst_wnd,vbox1,"Dismiss","cst_cancel",-1,-1,NULL);
	
	gtk_signal_connect (GTK_OBJECT (cst_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_cst_cancel_clicked), cst_wnd);
	gtk_signal_connect (GTK_OBJECT (cst_wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_cst_cancel_clicked), cst_wnd);
	gtk_widget_show (cst_wnd);
}








/******************************************************

	ARTIFACTS
	With these functions, we can spoil images with
	artifacts, such as noise etc.

******************************************************/


void add_gaussian_noise (image_type* img, float amplitude, int slice, int roiflg)
{
int x,y,z,i;
int c;
float f;
int bx1,by1,bx2,by2;

	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax; by2=img->ymax;
	}
	
	z = slice * img->xmax * mainimg.ymax;
	switch (mainimg.imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						c=(int) (amplitude * gauss_rnd() );	/* Note: Amplitude is stdev here */
						c = p[x + mainimg.xmax*y + z] + c;
						if (c<0) c=0; else if (c>255) c=255;
						p[x + mainimg.xmax*y + z] = c;
				}
			break;
		}
		case B_SHORT:
		{
			short* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						c=(int) (amplitude * gauss_rnd() );
						c = p[x + mainimg.xmax*y + z] + c;
						p[x + mainimg.xmax*y + z] = c;
					}
			break;
		}
		case B_RGB:
		{
			unsigned char* p = mainimg.data;
			for (i=0; i<3; i++)
				for (y=by1; y<=by2; y++)
					for (x=bx1; x<=bx2; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (amplitude * gauss_rnd() );
							c = p[i+3*(x + mainimg.xmax*y + z)] + c;
							if (c<0) c=0; else if (c>255) c=255;
							p[i+3*(x + mainimg.xmax*y + z)] = c;
						}
			break;
		}
		case B_FLOAT:
		{
			float* p = mainimg.data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						f= (amplitude * gauss_rnd() );
						f = p[x + mainimg.xmax*y + z] + f;
						p[x + mainimg.xmax*y + z] = f;
					}
			break;
		}
	}
	
	img->uid = next_uid++;

}


void add_shot_noise (image_type* img, float probability, int slice, int roiflg)
{
int x,y,z,i;
int c, ct;
float f,a,b;
int bx1,by1,bx2,by2;

	a = 200; b=-0.5*a;
	ct = a*0.01*probability; 	/* a good value for prob is 5 (5%) so that ct==10 */
	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax; by2=img->ymax;
	}

	if (file_ok) 
	{
		z = slice * img->xmax * img->ymax;
		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								c = img->maxval.imax;
								p[x + img->xmax*y + z] = c;
							}
							else if ((c<0) && (c>-ct))
							{
								c = img->minval.imin;
								p[x + img->xmax*y + z] = c;
							}

						}
				break;
			}
			case B_SHORT:
			{
				short* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								c = img->maxval.imax;
								p[x + img->xmax*y + z] = c;
							}
							else if ((c<0) && (c>-ct))
							{
								c = img->minval.imin;
								p[x + img->xmax*y + z] = c;
							}
						}
				break;
			}
			case B_RGB:
			{
				unsigned char* p = img->data;
				for (i=0; i<3; i++)
					for (y=0; y<img->ymax; y++)
						for (x=0; x<img->xmax; x++)
							if (in_roi(x,y)^roiflg)
							{
								c=(int) (b+a*rand()/(RAND_MAX+1.0));
								if ((c>0) && (c<ct))
								{
									c = img->maxval.imax;
									p[i + 3*(x + img->xmax*y + z)] = c;
								}
								else if ((c<0) && (c>-ct))
								{
									c = img->minval.imin;
									p[i + 3*(x + img->xmax*y + z)] = c;
								}
							}
				break;
			}
			case B_FLOAT:
			{
				float* p = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
						if (in_roi(x,y)^roiflg)
						{
							c=(int) (b+a*rand()/(RAND_MAX+1.0));
							if ((c>0) && (c<ct))
							{
								f = img->maxval.fmax;
								p[x + img->xmax*y + z] = f;
							}
							else if ((c<0) && (c>-ct))
							{
								f = img->minval.fmin;
								p[x + img->xmax*y + z] = f;
							}
						}
				break;
			}
		}

	}
	img->uid = next_uid++;

}


/* A rudiment related to the above code to add a line artifact 
   maybe we can use it somewhere?

	if ((x & 0x0f) ==1)
	{
		c=(int) (40 + 2*rand()/(RAND_MAX+1.0));
		p[x + img->xmax*y + z] += c;
	}

*/

void chg_image_values (image_type* img, float value, int slice, int roiflg)
{
int x,y,z,i;
int c;
int bx1,by1,bx2,by2;

	roi_bbox (img, &bx1,&by1, &bx2, &by2);
	c = (int)value;
	if (roiflg) 
	{
		roiflg=1;	/* Adjust to pure 0/1 */
		bx1=0; by1=0; bx2=img->xmax-1; by2=img->ymax-1;
	}
	
	z = slice * img->xmax * img->ymax;
	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
				{
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = c;
					}
				}
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = c;
					}
			break;
		}
		case B_RGB:
		{
			unsigned char* p = img->data;
			for (i=0; i<3; i++)
				for (y=by1; y<=by2; y++)
					for (x=bx1; x<=bx2; x++)
						if (in_roi(x,y)^roiflg)
						{
							p[i+3*(x + img->xmax*y + z)] = c;
						}
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			for (y=by1; y<=by2; y++)
				for (x=bx1; x<=bx2; x++)
					if (in_roi(x,y)^roiflg)
					{
						p[x + img->xmax*y + z] = value;
					}
			break;
		}
	}
	
	img->uid = next_uid++;

	
}




/******************************************************

	GUI for the Edit -> Values function

*******************************************************/

GtkAdjustment* edval_value_adj;
int edval_roi, edval_function, edval_allslices;
GtkWidget *edval_menu;


void on_edval_cancel_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_edval_ok_clicked (GtkButton *button, gpointer user_data)
{
float min, max;
int z,z1,z2;
gchar res;
	

	undo_nexus (&mainimg);

	imgminmax (&mainimg, &min, &max);
	if (edval_allslices)
	{
		z1=0; z2=mainimg.zmax-1;
	}
	else
		z1=z2=cur_slice;

	switch (edval_function)
	{
		case 0:	
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, 0, z, edval_roi);
			break;
		case 1:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, max, z, edval_roi);
			break;
		case 2:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, min, z, edval_roi);
			break;
		case 3:
			for (z=z1; z<=z2; z++)
				chg_image_values (&mainimg, edval_value_adj->value, z, edval_roi);
			break;
		case 4:
			for (z=z1; z<=z2; z++)
				add_gaussian_noise (&mainimg, edval_value_adj->value, z, edval_roi);
			break;
		case 5:
			for (z=z1; z<=z2; z++)
				add_shot_noise (&mainimg, edval_value_adj->value, z, edval_roi);
	}
	
	make_minmax (&mainimg);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);
	
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_edval_checkbutton1_toggled (GtkButton *button, gpointer user_data)
{
	edval_roi = (*(GtkToggleButton*)user_data).active;

}


void on_edval_checkbutton2_toggled (GtkButton *button, gpointer user_data)
{
	edval_allslices = (*(GtkToggleButton*)user_data).active;

}


void edval_set_menu (GtkWidget *widget, gpointer user_data)
{

	edval_function = *(int*)user_data;
	
}

void on_edval_value_changed (GtkWidget *widget, gpointer user_data)
{

	edval_function = menuopt3;
	gtk_option_menu_set_history (GTK_OPTION_MENU (edval_menu), edval_function);
	
}



void create_editval_wnd (image_type* img)
{
  GtkWidget *edval_wnd;
  GtkWidget* vbox1;
  GtkWidget *table1;
  GtkWidget *edval_label1;
  GtkWidget *edval_label2;
  GtkWidget *edval_label3;
  GtkWidget *edval_label4;
  GtkWidget *edval_roisel;
  GtkWidget *edval_slicesel;
  GtkWidget *edval_cancel;
  GtkWidget *edval_ok;
  GtkWidget *edval_value;
  GtkWidget* hsep1;
  GtkWidget* hsep2;
  GtkWidget *edval_menu_menu;
  GtkWidget *glade_menuitem;
 
  float min, max;
  
  	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
			min = 0; max = 255;
			break;
		case B_SHORT:
			min = -32767; max = 32767;
			break;
		case B_FLOAT:
			min = MIN(img->minval.fmin,-65535); max = MAX(img->maxval.fmax,65535);
			break;
		default:
			min=0; max=0; 	/* kill a compiler warning */
	}

	edval_roi=0;	edval_allslices=0; edval_function=0;		

	edval_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (edval_wnd), "edval_wnd", edval_wnd);
	gtk_window_set_title (GTK_WINDOW (edval_wnd), "Edit Values");
	gtk_window_set_modal (GTK_WINDOW (edval_wnd), TRUE);

	vbox1 = vbox_in_window (edval_wnd, "edval_vbox1");
	table1 = MakeTable (edval_wnd, vbox1, "table1", 2,6);
	edval_label1 = MakeLabel (edval_wnd,table1,"Region","edval_label1",0,0);
	edval_label2 = MakeLabel (edval_wnd,table1,"Slice","edval_label2",0,1);
	edval_label3 = MakeLabel (edval_wnd,table1,"Function","edval_label3",0,2);
	edval_label4 = MakeLabel (edval_wnd,table1,"Value","edval_label4",0,3);

	edval_roisel = Checkbutton_in_table (edval_wnd, table1, "Outside ROI","edval_check1",1,0, NULL);
	edval_slicesel = Checkbutton_in_table (edval_wnd, table1, "All slices","edval_check2",1,1, NULL);
		
	edval_menu = Make_Optionmenu (edval_wnd,table1,"edval_menu",1,2);
	edval_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Zero", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Maximum", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Minimum", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Specified value", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Gaussian noise", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (edval_menu_menu, GTK_SIGNAL_FUNC(edval_set_menu), "Shot noise", (gpointer) &menuopt5);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (edval_menu), edval_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (edval_menu), 0);

	edval_value_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, min, max, 1, 1, 0));
	edval_value = MakeSpinbutton (edval_value_adj,edval_wnd,table1,"edval_value",1,3);
	if (img->imgtype==B_FLOAT) gtk_spin_button_set_digits (GTK_SPIN_BUTTON(edval_value), 1);

	hsep1 = MakeHsep (edval_wnd,table1,"hsep1",0,4);
	hsep2 = MakeHsep (edval_wnd,table1,"hsep2",1,4);
	
	edval_cancel = Button_in_table (edval_wnd,table1,"Cancel","edval_cancel",1,5,NULL);
	edval_ok = Button_in_table (edval_wnd,table1,"OK","edval_ok",0,5,NULL);

	gtk_signal_connect (GTK_OBJECT (edval_wnd), "destroy", GTK_SIGNAL_FUNC (on_edval_cancel_clicked), edval_wnd);
	gtk_signal_connect (GTK_OBJECT (edval_ok), "clicked", GTK_SIGNAL_FUNC (on_edval_ok_clicked), edval_wnd);
	gtk_signal_connect (GTK_OBJECT (edval_roisel), "toggled", GTK_SIGNAL_FUNC (on_edval_checkbutton1_toggled), edval_roisel);
	gtk_signal_connect (GTK_OBJECT (edval_value_adj), "value_changed", GTK_SIGNAL_FUNC (on_edval_value_changed), edval_value_adj);
	gtk_signal_connect (GTK_OBJECT (edval_slicesel), "toggled", GTK_SIGNAL_FUNC (on_edval_checkbutton2_toggled), edval_slicesel);
	gtk_signal_connect (GTK_OBJECT (edval_cancel), "clicked", GTK_SIGNAL_FUNC (on_edval_cancel_clicked), edval_wnd);

	gtk_widget_show (edval_wnd);
}


void editval_main ()
{

	if (!current_roi.roi_mode)
		ShowMessage ("This function requires a ROI");
	else
		create_editval_wnd (&mainimg);

}



/******************************************************************

	Preferences Dialog
	
******************************************************************/

struct 
{
	int tiffreport;
	int intp_prefs;
	int debuglevel;
	int skip_license;
	UNDO_TYPE undo_function;
	char browspath[1024];
	char helppath[1024];
	char dsopath[1024];
	char undo_path[1024];
	char imager_sv_path[1024];
}
prefs;

GtkWidget* prefs_entry1;
GtkWidget* prefs_entry2;
GtkWidget* prefs_entry3;
GtkWidget* prefs_entry4;
GtkWidget* prefs_entry5;


/* This function creates the initial Crystal Image .rc file. It is placed here because of context. */

void save_default_prefs ()
{
FILE* FP;
char fname [256];
char* home;

	prefs.tiffreport = 0;
	prefs.intp_prefs = INTP_BILIN;
	prefs.debuglevel = 1;
	prefs.skip_license=1;
	prefs.undo_function = 1;
	strcpy (prefs.browspath, "/usr/bin/konqueror");		/* TODO: Should query several browsers */
	strcpy (prefs.helppath, "/usr/local/cimage/doc");
	prefs.dsopath[0] = 0;								/* Keep empty b/c we look into /usr/local anyway */
	prefs.undo_path[0] = 0;
	//strcpy (prefs.imager_sv_path, "/usr/local/cimage/bin/cimage_sv"); /* Disabled because of Radeon problems */
	prefs.imager_sv_path[0]=0;


	home = getenv ("HOME");
	if (!home)
		ShowMessage ("Cannot determine home directory");
	dp (2, "Home directory is %s\n", home);

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	dp (2, "Saving preferences in %s\n",fname);
	FP = fopen (fname, "wt");
	if (!FP) ShowMessage ("Error saving preferences");
	
	fprintf (FP, "Tiffinfo %d\n",prefs.tiffreport);
	fprintf (FP, "Interpolation %d\n",prefs.intp_prefs);
	fprintf (FP, "Debuglevel %d\n",prefs.debuglevel);
	fprintf (FP, "Skip_license %d\n",prefs.skip_license);
	fprintf (FP, "Undo_function %d\n",prefs.undo_function);
	fprintf (FP, "Path_to_undo \n");
	fprintf (FP, "Path_to_Browser %s\n",prefs.browspath);
	fprintf (FP, "Path_to_Help %s\n",prefs.helppath);
	fprintf (FP, "Path_to_DSOs %s\n",prefs.dsopath);
	fprintf (FP, "sv_viewer_name %s\n",prefs.imager_sv_path);
	fclose (FP);
	dp (1, "Preferences saved.\n");
}




void save_prefs ()
{
FILE* FP;
char fname [256];
char* home;


	home = getenv ("HOME");
	if (!home)
		ShowMessage ("Cannot determine home directory");
	dp (2, "Home directory is %s\n", home);

	strcpy (fname, home);
	strcat (fname, "/.cimagerc");
	dp (2, "Saving preferences in %s\n",fname);
	FP = fopen (fname, "wt");
	if (!FP) ShowMessage ("Error saving preferences");
	
	fprintf (FP, "Tiffinfo %d\n",tiffreport);
	fprintf (FP, "Interpolation %d\n",intp_pref);
	fprintf (FP, "Debuglevel %d\n",debuglevel);
	fprintf (FP, "Skip_license %d\n",skip_license);
	fprintf (FP, "Undo_function %d\n",undo_function);
	fprintf (FP, "Path_to_undo %s\n",undo_path);
	fprintf (FP, "Path_to_Browser %s\n",browspath);
	fprintf (FP, "Path_to_Help %s\n",helppath);
	fprintf (FP, "Path_to_DSOs %s\n",dsopath);
	fprintf (FP, "sv_viewer_name %s\n",imager_sv_path);
	fclose (FP);
	dp (1, "Preferences saved.\n");
}



void on_interp_menu_changed (GtkWidget *widget, gpointer user_data)
{
	prefs.intp_prefs = *(int*) user_data;
	dp (2, "Interpolation method set to %d\n",prefs.intp_prefs);
}


void on_undo_menu_changed (GtkWidget *widget, gpointer user_data)
{
	prefs.undo_function = *(int*) user_data;
	dp (2, "Undo function set to %d\n",prefs.undo_function);
}


void on_prefs_tiffbutton_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.tiffreport = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Tiffreport set to %d\n",prefs.tiffreport);
}


void on_prefs_licensebutton_clicked (GtkToggleButton *togglebutton, gpointer user_data)
{
	prefs.skip_license = gtk_toggle_button_get_active ((GtkToggleButton*) user_data);
	dp (2, "Skip-license set to %d\n",prefs.skip_license);
}


void on_debug_spinner_changed (GtkEditable *editable, gpointer user_data)
{
GtkAdjustment* adj;

	adj = user_data;
	prefs.debuglevel = adj->value;
	dp (2, "Debuglevel set to %d\n",prefs.debuglevel);
}

void on_prefs_cancelbutton_clicked ( GtkButton * button, gpointer data )
{
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void on_prefs_okbutton_clicked ( GtkButton * button, gpointer data )
{
	tiffreport=prefs.tiffreport;
	intp_pref=prefs.intp_prefs;
	debuglevel=prefs.debuglevel;
	skip_license = prefs.skip_license;
	
	if (undo_function != prefs.undo_function) undo_stack_invalidate ();
	undo_function=prefs.undo_function;

	memset (prefs.browspath,0,1020);
	strncpy (prefs.browspath,gtk_entry_get_text (GTK_ENTRY(prefs_entry1)),1000);
	strcpy (browspath,prefs.browspath);

	memset (prefs.helppath,0,1020);
	strncpy (prefs.helppath,gtk_entry_get_text (GTK_ENTRY(prefs_entry2)),1000);
	strcpy (helppath,prefs.helppath);

	memset (prefs.dsopath,0,1020);
	strncpy (prefs.dsopath,gtk_entry_get_text (GTK_ENTRY(prefs_entry3)),1000);
	strcpy (dsopath,prefs.dsopath);

	memset (prefs.undo_path,0,1020);
	strncpy (prefs.undo_path,gtk_entry_get_text (GTK_ENTRY(prefs_entry4)),1000);
	strcpy (undo_path,prefs.undo_path);
	
	memset (prefs.imager_sv_path,0,1020);
	strncpy (prefs.imager_sv_path,gtk_entry_get_text (GTK_ENTRY(prefs_entry5)),1000);
	strcpy (imager_sv_path,prefs.imager_sv_path);

	/* Save the new preferences */
	
	save_prefs ();

	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void create_prefswnd (void)
{
  GtkWidget *prefswnd;
  GtkWidget *vbox1;
  GtkWidget *table1;
  GtkWidget *label1, *label2, *label3, *label4, *label5, *label6, *label7, *label8, *label9, *label10;
  GtkWidget *hseparator1, *hseparator2;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  GtkWidget *tiffinfobutton;
  GtkWidget *licensebutton;
  GtkWidget *interp_menu;
  GtkWidget *interp_menu_menu;
  GtkWidget *undo_menu;
  GtkWidget *undo_menu_menu;
  GtkWidget *glade_menuitem;
  GtkObject *debug_spinner_adj;
  GtkWidget *debug_spinner;
  int y;

	prefswnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (prefswnd), "prefswnd", prefswnd);
	gtk_window_set_title (GTK_WINDOW (prefswnd), "Preferences");

	vbox1 = vbox_in_window (prefswnd, "vbox1");
	table1 = MakeTable (prefswnd, vbox1, "table1",2,9);
	y=0;

	label1 = MakeLabel (prefswnd,table1,"Show file import info","label1",0,y);
	tiffinfobutton = Checkbutton_in_table (prefswnd,table1," ","prefs_tiffinfo",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (tiffinfobutton), tiffreport);
	gtk_signal_connect (GTK_OBJECT (tiffinfobutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_tiffbutton_clicked), tiffinfobutton);
  	y++;

	label2 = MakeLabel (prefswnd,table1,"Preferred interpolation method","label2",0,y);
	interp_menu = Make_Optionmenu (prefswnd,table1,"interp_menu",1,y);
	interp_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Nearest Neighbor",&menuopt0);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Bilinear",&menuopt1);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Bicubic",&menuopt2);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (2nd order)",&menuopt3);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (3rd order)",&menuopt4);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (4th order)",&menuopt5);
	glade_menuitem = Option_menuitem (interp_menu_menu,on_interp_menu_changed,"Spline (5th order)",&menuopt6);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (interp_menu), interp_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (interp_menu), intp_pref);
	y++;

	label3 = MakeLabel (prefswnd,table1,"Debug message level (0-3)","label3",0,y);
	debug_spinner_adj = gtk_adjustment_new (debuglevel, 0, 3, 1, 1, 0);
	debug_spinner = MakeSpinbutton (GTK_ADJUSTMENT (debug_spinner_adj),prefswnd,table1,"prefsspin1",1,y);
	gtk_signal_connect (GTK_OBJECT (debug_spinner), "value-changed", GTK_SIGNAL_FUNC (on_debug_spinner_changed),debug_spinner_adj);
	y++;

	label8 = MakeLabel (prefswnd,table1,"Undo function","label8",0,y);
	undo_menu = Make_Optionmenu (prefswnd,table1,"undo_menu",1,y);
	undo_menu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO disabled",&menuopt0);
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO in memory",&menuopt1);
	glade_menuitem = Option_menuitem (undo_menu_menu,on_undo_menu_changed,"UNDO stack on disk",&menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (undo_menu), undo_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (undo_menu), undo_function);
	y++;

	label7 = MakeLabel (prefswnd,table1,"Temp path for undo","label7",0,y);
	prefs_entry4 = Entry_in_table (prefswnd,table1,prefs.undo_path,"prefs_entry4",1,y);
	y++;

	label4 = MakeLabel (prefswnd,table1,"Path to browser","label4",0,y);
	prefs_entry1 = Entry_in_table (prefswnd,table1,prefs.browspath,"prefs_entry1",1,y);
	y++;
	label5 = MakeLabel (prefswnd,table1,"Path to help files","label5",0,y);
	prefs_entry2 = Entry_in_table (prefswnd,table1,prefs.helppath,"prefs_entry2",1,y);
	y++;
	label6 = MakeLabel (prefswnd,table1,"Path to DSO modules","label6",0,y);
	prefs_entry3 = Entry_in_table (prefswnd,table1,prefs.dsopath,"prefs_entry3",1,y);
	y++;
	label9 = MakeLabel (prefswnd,table1,"Path+name of 3D viewer","label9",0,y);
	prefs_entry5 = Entry_in_table (prefswnd,table1,prefs.imager_sv_path,"prefs_entry5",1,y);
	y++;

	label10 = MakeLabel (prefswnd,table1,"Suppress license splash","label10",0,y);
	licensebutton = Checkbutton_in_table (prefswnd,table1," ","prefs_license",1,y,NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (licensebutton), skip_license);
	gtk_signal_connect (GTK_OBJECT (licensebutton), "toggled", GTK_SIGNAL_FUNC (on_prefs_licensebutton_clicked), licensebutton);
	y++;

	hseparator1 = MakeHsep (prefswnd,table1,"hsep1",0,y);
	hseparator2 = MakeHsep (prefswnd,table1,"hsep2",1,y);
	y++;

	okbutton = Button_in_table (prefswnd,table1,"OK","prefs_ok",0,y,on_prefs_okbutton_clicked);
	cancelbutton = Button_in_table (prefswnd,table1,"Cancel","prefs_cancel",1,y,on_prefs_cancelbutton_clicked);

	gtk_widget_show (prefswnd);
	
}


void preferences_main ()
{

	prefs.tiffreport=tiffreport;
	prefs.skip_license=skip_license;
	prefs.intp_prefs=intp_pref;
	prefs.debuglevel=debuglevel;
	prefs.undo_function = undo_function;
	strcpy (prefs.dsopath, dsopath);
	strcpy (prefs.browspath,browspath);
	strcpy (prefs.helppath,helppath);
	strcpy (prefs.undo_path, undo_path);
	strcpy (prefs.imager_sv_path, imager_sv_path);
	create_prefswnd ();
}


	

/*************************************************************

	Karhunen-Loeve-Transform (Principal component analysis)

	The K-L transform acts on RGB images by interpreting
	each triplet as a point in three-dimensional RGB space.
	The K-L transform performs a coordinate system trans-
	formation so that one axis has the highest contrast
	(the principal component). The principal component
	is then returned as grayscale (luminance) value.

*************************************************************/


/* Compute covariance matrix in C */

void kl_comp_covar_mat (image_type* img, MFLOAT* C, int slice, int oorflg)
{
MFLOAT mr,mg,mb,sum;
int i,j;
long l,z,cnt;
unsigned char* p;
rgbtriplet* prgb;
rgbtriplet rgb;


	g_assert (img->imgtype==B_RGB);
	p = img->data; prgb=img->data;
	
	/* Step 1: Compute the image mean for each pane */
	
	z = slice * img->xmax * img->ymax;
	mr=0.0; mg=0.0; mb=0.0; cnt=0;
	for (l=0; l<img->xmax * img->ymax; l++)
	{
		rgb = prgb[l+z];
		if ((!oorflg) || (rgb.red>0) || (rgb.green>0) || (rgb.blue>0))
		{
			mr += rgb.red; mg += rgb.green; mb += rgb.blue;
			cnt++;
		}
	}
	if (cnt>0)
	{
		mr/=cnt; mb/=cnt; mg/=cnt;
	}
	else
		ShowMessage ("Cannot compute image mean - no data?");


	/* Step 2: Compute the covariance matrix */
		
	cnt=0; sum=0.0;
	for (l=0; l<img->xmax * img->ymax; l++)
	{
		rgb = prgb[l+z];
		if ((!oorflg) || (rgb.red>0) || (rgb.green>0) || (rgb.blue>0))
		{
			C[0] += (rgb.red   - mr)*(rgb.red   - mr);	/* RR = 0,0 */
			C[1] += (rgb.red   - mr)*(rgb.green - mg);	/* RG = 0,1 */
			C[2] += (rgb.red   - mr)*(rgb.blue  - mb);	/* RB = 0,2 */
			C[4] += (rgb.green - mg)*(rgb.green - mg);	/* GG = 1,1 */
			C[5] += (rgb.green - mg)*(rgb.blue  - mg);	/* GB = 1,2 */
			C[8] += (rgb.blue  - mg)*(rgb.blue  - mg);	/* BB = 2,2 */
			cnt++;
		}
	}
	
	C[3] = C[1];		/* RG = GR */
	C[7] = C[5];		/* GB = BG */
	C[6] = C[2];		/* BR = RB */
	for (i=0; i<9; i++) C[i] /= cnt;
	
	dp (2, "K-L Transform: This is the covariance matrix:\n");
	for (i=0; i<3; i++)
	{
		for (j=0; j<3; j++)
			dp (2, "%7.2f   ",C[3*i+j]);
		dp (2, "\n");
	}
	dp (2,"----------------------------------------\n");
}



/* Step 3: Compute the eigenvectors and the transposed and sorted
   eigenmatrix */

void kl_comp_transform_mat (MFLOAT* C)
{
MFLOAT D[3], E[3];
int i,j;

	householder_reduction (C,3,D,E);
	eigen_ql (C,3,D,E);
	
	dp (2, "K-L Transform: Eigenvectors ");
	for (i=0; i<3; i++) dp (2, "%f   ",D[i]);
	dp (2,"\n");
	dp (2, "K-L Transform: This is the transformation matrix:\n");
	for (i=0; i<3; i++)
	{
		for (j=0; j<3; j++)
			dp (2, "%7.2f   ",C[3*i+j]);
		dp (2, "\n");
	}
	dp (2,"----------------------------------------\n");

}


void karhunen_loeve_transform (image_type* img, int oorflg)
{
MFLOAT C[9];
image_type buf;
MFLOAT V[3];
int x,y,z;
long zidx;
rgbtriplet* p1;
float* p2;
rgbtriplet rgb;


	g_assert (img->imgtype==B_RGB);
	
	/* Allocate the output buffer (float grayscale) */

	if (allocate_image (&buf, B_FLOAT, img->xmax, img->ymax, img->zmax) ==-1)
		fatal_memerr();
	p1 = img->data;
	p2 = buf.data;

	/* Transform slice by slice (TODO: is that a good idea?? Or should we
	   compute one transformation matrix for the entire stack?) */

	for (z=0; z<img->zmax; z++)
	{
		dp (1, "Karhunen-Loeve-Transform: Processing slice %d\n",z);

		kl_comp_covar_mat (img, C, z, oorflg);
		kl_comp_transform_mat (C);

		zidx = z*img->xmax*img->ymax;
		
		/* Pixel-by-pixel conversion */
		
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				rgb = p1[x+img->xmax*y+zidx];
				V[0] = (float)rgb.red;
				V[1] = (float)rgb.green;
				V[2] = (float)rgb.blue;
				matmult (C,V,3);
				p2[x+img->xmax*y+zidx] = V[0]; 	/* only principal component */
			}

	}
	
	/* return the new b&w image */
	
	handback (img, &buf);

}
	
	


/*******************************************************************/

/*	A general-purpose 2D flood-fill function. It works as follows:
		- x and y are the seed values in slice z
		- newval is the new value to flood-fill with
		- level needs to be passed (initially level=0) to keep track of the recursion level

		2D flood fill distinguishes two conditions. If newval is zero, 
			it flood-fills a contigugous region with zeros, i.e., makes a cluster vanish
		If newval is >0, it flood-fills a contigous region of zero-values pixels with 
			the new value (filling holes, for example)
*/

void rg_floodfill (image_type* img, float newval, int x, int y, int z, int* level)
{

	if (*level==0) dp (1, "Flood-filling from (%d,%d) in slice %d with value %f\n",x,y,z,newval);

	/* Abort recursion if out-of-range or non-background or already visited */
	
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) return;

	/* Abort recursion if
		1) newval is zero and the current pixel is also zero
		2) newval is nonzero and the current pixel is nonzero */

	if (F_ISZERO(newval))
	{
		if (F_ISZERO(readbuf_flt (img,x,y,z))) return;
	}
	else
	{
		if (!F_ISZERO(readbuf_flt (img,x,y,z))) return;
	}

	
	(*level)++;
	
	/* Set the current point to newval */

	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char *p = img->data;
			int c;
			if (newval<0) c=0; else if (newval>255) c=255; else c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]=c;
			break;
		}
		case B_SHORT:
		{
			short *p = img->data;
			int c;
			c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]=c;
			break;
		}
		case B_FLOAT:
		{
			float *p = img->data;
			p[x + img->xmax*(y + img->ymax*z)]=newval;
			break;
		}
		case B_RGB:
		{
			rgbtriplet *p = img->data;
			int c;
			if (newval<0) c=0; else if (newval>255) c=255; else c=IROUND(newval);
			p[x + img->xmax*(y + img->ymax*z)]= (rgbtriplet){c,c,c};
			break;
		}
	}


	/* and visit the 8 neighbors */
	
	rg_floodfill (img,newval,x+1,y,z,level);
	rg_floodfill (img,newval,x-1,y,z,level);
	rg_floodfill (img,newval,x,y+1,z,level);
	rg_floodfill (img,newval,x,y-1,z,level);
	rg_floodfill (img,newval,x+1,y+1,z,level);
	rg_floodfill (img,newval,x-1,y+1,z,level);
	rg_floodfill (img,newval,x+1,y-1,z,level);
	rg_floodfill (img,newval,x-1,y-1,z,level);
	(*level)--;
}




/*******************************************************************/
